/*
 * Copyright 2008 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.derquinsej.collect;

import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.base.Function;
import com.google.common.base.Preconditions;
import com.google.common.base.Supplier;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;

/**
 * An immutable and thread-safe implementation of hierarchy.
 * @author Andres Rodriguez
 * @param <K> Type of the keys.
 * @param <V> Type of the elements.
 */
public final class ImmutableHierarchy<K, V> implements Hierarchy<K, V> {
	private final Function<? super V, K> keyFunction;
	private final ImmutableMap<K, V> elements;
	private final ImmutableList<V> firstLevel;
	private final ImmutableMap<K, V> parents;
	private final ImmutableMultimap<K, V> children;

	private ImmutableHierarchy(final Builder<K, V> builder) {
		this.keyFunction = builder.keyFunction;
		this.elements = ImmutableMap.copyOf(builder.elements);
		this.firstLevel = ImmutableList.copyOf(builder.firstLevel);
		final ImmutableMap.Builder<K, V> pb = ImmutableMap.builder();
		for (final Map.Entry<K, K> pe : builder.parents.entrySet()) {
			pb.put(pe.getKey(), this.elements.get(pe.getValue()));
		}
		this.parents = pb.build();
		this.children = builder.children.build();
	}

	/**
	 * Returns a new builder.
	 */
	public static <K, V> Builder<K, V> builder(final Function<? super V, K> keyFunction) {
		return new Builder<K, V>(keyFunction);
	}

	private static void checkKey(Object key) {
		Preconditions.checkNotNull(key, "Elements with a null key are not allowed");
	}

	private static <K, V> K getKey(V element, Function<? super V, K> keyFunction) {
		Preconditions.checkNotNull(element, "Null elements are not allowed");
		final K key = keyFunction.apply(element);
		checkKey(key);
		return key;
	}

	private K getMemberKey(V element) {
		final K key = getKey(element, keyFunction);
		Preconditions.checkArgument(elements.containsKey(key), "The provided element is not part of the hierarchy");
		return key;
	}

	private void checkMemberKey(K key) {
		checkKey(key);
		Preconditions.checkArgument(elements.containsKey(key),
				"There is no element with the provided key in the hierarchy");
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.collect.Hierarchy#getFirstLevel()
	 */
	public ImmutableList<V> getFirstLevel() {
		return firstLevel;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.collect.Hierarchy#getChildren(java.lang.Object)
	 */
	public ImmutableList<V> getChildren(V element) {
		return children.get(getMemberKey(element));
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.collect.Hierarchy#getParent(java.lang.Object)
	 */
	public V getParent(V element) {
		return parents.get(getMemberKey(element));
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.collect.Hierarchy#contains(java.lang.Object)
	 */
	public boolean contains(V element) {
		return elements.containsKey(getKey(element, keyFunction));
	}

	/*
	 * (non-Javadoc)
	 * @see
	 * net.sf.derquinsej.collect.Hierarchy#getChildrenByKey(java.lang.Object)
	 */
	public ImmutableList<V> getChildrenByKey(K key) {
		checkMemberKey(key);
		return children.get(key);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.collect.Hierarchy#getParentByKey(java.lang.Object)
	 */
	public V getParentByKey(K key) {
		checkMemberKey(key);
		return parents.get(key);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.collect.Hierarchy#containsKey(java.lang.Object)
	 */
	public boolean containsKey(K key) {
		checkKey(key);
		return elements.containsKey(key);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.collect.Hierarchy#get(java.lang.Object)
	 */
	public V get(K key) {
		checkKey(key);
		return elements.get(key);
	}

	/**
	 * Builder for immutable hierarchies.
	 * @author Andres Rodriguez
	 * @param <K> Type of the keys.
	 * @param <V> Type of the elements.
	 */
	public static final class Builder<K, V> implements Supplier<ImmutableHierarchy<K, V>> {
		private final Function<? super V, K> keyFunction;
		private final Map<K, V> elements = Maps.newHashMap();
		private final Set<K> unaddedParents = Sets.newHashSet();
		private final List<V> firstLevel = Lists.newLinkedList();
		private final Map<K, K> parents = Maps.newHashMap();
		private final ImmutableMultimap.Builder<K, V> children = ImmutableMultimap.builder();

		private Builder(final Function<? super V, K> keyFunction) {
			Preconditions.checkNotNull(keyFunction);
			this.keyFunction = keyFunction;
		}

		/**
		 * Adds a new element to the hierarchy.
		 * @param element Element to add.
		 * @param parentKey Key of the parent element, or {@code null} if it is
		 *            a first level node.
		 * @return This builder.
		 * @throws NullPointerException if element is {@code null}.
		 * @throws IllegalArgumentException if the node has been previously
		 *             added.
		 * @throws IllegalStateException if a loop is detected.
		 */
		public Builder<K, V> add(V element, K parentKey) {
			final K key = getKey(element, keyFunction);
			Preconditions.checkArgument(!elements.containsKey(key), "Duplicate nodes not allowed");
			if (parentKey == null) {
				firstLevel.add(element);
			} else {
				K p = parentKey;
				final Set<K> visited = Sets.newHashSet();
				visited.add(key);
				while (p != null) {
					Preconditions.checkState(visited.add(p), "Loop detected");
					p = parents.get(p);
				}
				children.put(parentKey, element);
				parents.put(key, parentKey);
				if (!elements.containsKey(parentKey)) {
					unaddedParents.add(parentKey);
				}
			}
			unaddedParents.remove(key);
			elements.put(key, element);
			return this;
		}

		/**
		 * Builds and returns an immutable hierarchy with the nodes added up to
		 * the method call.
		 * @returns An immutable hierarchy.
		 * @throws IllegalStateException if there are referenced parents that
		 *             are not part of the hierarchy yet.
		 */
		public ImmutableHierarchy<K, V> get() {
			Preconditions.checkState(unaddedParents.isEmpty(),
					"There are referenced parents that have not been added yet.");
			return new ImmutableHierarchy<K, V>(this);
		}

	}

}
